require 'active_support/tagged_logging'
require 'active_support/logger'

module Dronejob
  module Modules
    module Log
      extend ActiveSupport::Concern
      COLOR_BOLD = "\e[1m"
      COLOR_RED = "\e[31m"
      COLOR_RED_BG = "\e[41m"
      COLOR_WHITE = "\e[38;5;15m"
      COLOR_GREY = "\e[38;5;245m"
      COLOR_GREY_BG = "\e[48;5;255m"
      COLOR_END = "\e[0m"

      included do
        cattr_accessor(:logger) do
          dronejob_log = option(:dronejob_log) || "STDOUT"
          case dronejob_log
          when "STDOUT"
            custom_logger = ActiveSupport::Logger.new(STDOUT)
          when "google"
            logging = Google::Cloud::Logging.new(project: option(:google_cloud_project_id))
            resource = logging.resource("gce_instance", { zone: option(:google_cloud_zone), instance_id: option(:google_cloud_instance_id) })
            custom_logger = logging.logger(option(:dronejob_queue), resource)
          else
            custom_logger = ActiveSupport::Logger.new(dronejob_log)
          end
          ActiveSupport::TaggedLogging.new(custom_logger)
        end
      end

      class_methods do
        def log(level, message, job = nil)
          if !logger.nil?
            if option(:dronejob_log) == "google"
              message = "[#{job&.job_id}] #{message}" if level != :error
              data = { message: message, job_id: job&.job_id, queue: job&.queue_name, phase: "none" }
              data[:phase_title] = job.phase_config[:title].to_s if job&.phase_config&.key?(:title)
              if job&.phase_config&.key?(:progress)
                data[:phase_progress] = job.phase_config[:progress]
              end
              logger.send(level, data.compact)
            else
              logger.send(level, message)
            end
          elsif job
            puts "#{level}: [#{job.phase}] #{message}"
          else
            puts "#{level}: #{message}"
          end
        end
      end

      def log(level, message)
        self.class.log(level, message, self)
      end

      def debug(message)
        log(:debug, message)
      end

      def info(message)
        log(:info, message)
      end

      def notice(message)
        log(:notice, message)
      end

      def warning(message)
        log(:warning, message)
      end

      def error(message)
        if message.kind_of?(StandardError)
          if option(:dronejob_log) == "google"
            log(:error, "#{message.backtrace.first}: #{message.message} (#{message.class})\n\t#{message.backtrace.drop(1).join("\n\t")}")
          else
            if log_stdout?
              log(:error, "#{COLOR_RED_BG}#{COLOR_BOLD}#{COLOR_WHITE} ERROR #{COLOR_END} #{COLOR_BOLD}#{message.class}: #{COLOR_RED}#{message.message}#{COLOR_END}")
            else
              log(:error, "[ERROR] #{message.class}: #{message.message}")
            end
            backtrace(message.backtrace)
          end
        else
          log(:error, message)
        end
      end

      def backtrace(lines)
        lines.each do |line|
          line.gsub!("#{Dir.pwd}/", '')
          (file, lineno, info) = line.split(":")
          if %r{ruby/?gems}.match(line).nil?
            if log_stdout?
              log(:error, "#{COLOR_GREY_BG}#{COLOR_GREY} #{lineno.rjust(5, ' ')} #{COLOR_END} #{COLOR_RED}#{file}#{COLOR_END} #{COLOR_GREY}#{info}#{COLOR_END}")
            else
              log(:error, " - #{file}:#{lineno} #{info}")
            end
          end
        end
      end

      def error!(message)
        throw message
      end

      def critical(message)
        log(:critical, message)
      end

      def alert(message)
        log(:alert, message)
      end

      def emergency(message)
        log(:emergency, message)
      end

      ActiveSupport.on_load :active_job do
        class ActiveJob::Logging::LogSubscriber
          def enqueue(event)
            info do
              job = event.payload[:job]
              log_object("queueing #{job.class.name}", job)
            end
          end

          def enqueue_at(event)
            info do
              job = event.payload[:job]
              log_object("scheduling #{job.class.name}", job)
            end
          end

          def perform_start(event)
            info do
              job = event.payload[:job]
              log_object("starting #{job.class.name}", job)
            end
          end

          def perform(event)
            info do
              job = event.payload[:job]
              log_object("finishing #{job.class.name}", job)
            end
          end

          def log_object(message, job)
            if job.option(:dronejob_log) == "google"
              { message: "[#{job.job_id}] #{message}", job_id: job.job_id, queue: job.queue_name, phase: job.phase }
            else
              message
            end
          end
        end
      end

      private

      def log_stdout?
        logdev = logger.instance_variable_get("@logdev")
        logger_source = logdev.dev if logdev.respond_to?(:dev)
        STDOUT == logger_source
      end
    end
  end
end
